@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2008 Zope Corporation and Contributors.
+# Copyright (c) 2008 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -15,6 +15,8 @@
 
 import logging
 import re
+import time
+
 import cx_Oracle
 from ZODB.POSException import StorageError
 
@@ -223,9 +225,10 @@ class OracleAdapter(Adapter):
         -- the object and all its revisions will be removed.
         -- If keep is 'Y', instead of removing the object,
         -- the pack operation will cut the object's history.
-        -- If keep is 'Y' then the keep_tid field must also be set.
-        -- The keep_tid field specifies which revision to keep within
-        -- the list of packable transactions.
+        -- The keep_tid field specifies the oldest revision
+        -- of the object to keep.
+        -- The visited flag is set when pre_pack is visiting an object's
+        -- references, and remains set.
         CREATE TABLE pack_object (
             zoid        NUMBER(20) NOT NULL PRIMARY KEY,
             keep        CHAR NOT NULL CHECK (keep IN ('N', 'Y')),
@@ -250,7 +253,8 @@ class OracleAdapter(Adapter):
         -- Temporary state during packing: a list of objects
         -- whose references need to be examined.
         CREATE GLOBAL TEMPORARY TABLE temp_pack_visit (
-            zoid        NUMBER(20) NOT NULL PRIMARY KEY
+            zoid        NUMBER(20) NOT NULL PRIMARY KEY,
+            keep_tid    NUMBER(20)
         );
 
         -- Temporary state during undo: a list of objects
@@ -261,55 +265,52 @@ class OracleAdapter(Adapter):
         );
         """
         self._run_script(cursor, stmt)
+        # Let Oracle catch up with the new data definitions by sleeping.
+        # This reduces the likelihood of spurious ORA-01466 errors.
+	time.sleep(5)
 
 
     def prepare_schema(self):
         """Create the database schema if it does not already exist."""
-        conn, cursor = self.open()
-        try:
-            try:
-                cursor.execute("""
-                SELECT 1 FROM USER_TABLES WHERE TABLE_NAME = 'OBJECT_STATE'
-                """)
-                if not cursor.fetchall():
-                    self.create_schema(cursor)
-            except:
-                conn.rollback()
-                raise
-            else:
-                conn.commit()
-        finally:
-            self.close(conn, cursor)
-
+        def callback(conn, cursor):
+            cursor.execute("""
+            SELECT 1 FROM USER_TABLES WHERE TABLE_NAME = 'OBJECT_STATE'
+            """)
+            if not cursor.fetchall():
+                self.create_schema(cursor)
+        self._open_and_call(callback)
 
     def zap_all(self):
         """Clear all data out of the database."""
-        conn, cursor = self.open()
-        try:
-            try:
-                stmt = """
-                DELETE FROM object_refs_added;
-                DELETE FROM object_ref;
-                DELETE FROM current_object;
-                DELETE FROM object_state;
-                DELETE FROM transaction;
-                -- Create a transaction to represent object creation.
-                INSERT INTO transaction (tid, username, description) VALUES
-                    (0, UTL_I18N.STRING_TO_RAW('system', 'US7ASCII'),
-                    UTL_I18N.STRING_TO_RAW(
-                    'special transaction for object creation', 'US7ASCII'));
-                DROP SEQUENCE zoid_seq;
-                CREATE SEQUENCE zoid_seq;
-                """
-                self._run_script(cursor, stmt)
-            except:
-                conn.rollback()
-                raise
-            else:
-                conn.commit()
-        finally:
-            self.close(conn, cursor)
+        def callback(conn, cursor):
+            stmt = """
+            DELETE FROM object_refs_added;
+            DELETE FROM object_ref;
+            DELETE FROM current_object;
+            DELETE FROM object_state;
+            DELETE FROM transaction;
+            -- Create a transaction to represent object creation.
+            INSERT INTO transaction (tid, username, description) VALUES
+                (0, UTL_I18N.STRING_TO_RAW('system', 'US7ASCII'),
+                UTL_I18N.STRING_TO_RAW(
+                'special transaction for object creation', 'US7ASCII'));
+            DROP SEQUENCE zoid_seq;
+            CREATE SEQUENCE zoid_seq;
+            """
+            self._run_script(cursor, stmt)
+        self._open_and_call(callback)
 
+    def drop_all(self):
+        """Drop all tables and sequences."""
+        def callback(conn, cursor):
+            for tablename in ('pack_state_tid', 'pack_state',
+                    'pack_object', 'object_refs_added', 'object_ref',
+                    'current_object', 'object_state', 'transaction',
+                    'commit_lock', 'pack_lock',
+                    'temp_store', 'temp_undo', 'temp_pack_visit'):
+                cursor.execute("DROP TABLE %s" % tablename)
+            cursor.execute("DROP SEQUENCE zoid_seq")
+        self._open_and_call(callback)
 
     def open(self, transaction_mode="ISOLATION LEVEL READ COMMITTED",
             twophase=False):
