@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2008 Zope Corporation and Contributors.
+# Copyright (c) 2008 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -106,8 +106,8 @@ class PostgreSQLAdapter(Adapter):
         -- the object and all its revisions will be removed.
         -- If keep is true, instead of removing the object,
         -- the pack operation will cut the object's history.
-        -- The keep_tid field specifies which revision to keep within
-        -- the list of packable transactions.
+        -- The keep_tid field specifies the oldest revision
+        -- of the object to keep.
         -- The visited flag is set when pre_pack is visiting an object's
         -- references, and remains set.
         CREATE TABLE pack_object (
@@ -142,49 +142,45 @@ class PostgreSQLAdapter(Adapter):
 
     def prepare_schema(self):
         """Create the database schema if it does not already exist."""
-        conn, cursor = self.open()
-        try:
-            try:
-                cursor.execute("""
-                SELECT tablename
-                FROM pg_tables
-                WHERE tablename = 'object_state'
-                """)
-                if not cursor.rowcount:
-                    self.create_schema(cursor)
-            except:
-                conn.rollback()
-                raise
-            else:
-                conn.commit()
-        finally:
-            self.close(conn, cursor)
-
+        def callback(conn, cursor):
+            cursor.execute("""
+            SELECT tablename
+            FROM pg_tables
+            WHERE tablename = 'object_state'
+            """)
+            if not cursor.rowcount:
+                self.create_schema(cursor)
+        self._open_and_call(callback)
 
     def zap_all(self):
         """Clear all data out of the database."""
-        conn, cursor = self.open()
-        try:
-            try:
-                cursor.execute("""
-                DELETE FROM object_refs_added;
-                DELETE FROM object_ref;
-                DELETE FROM current_object;
-                DELETE FROM object_state;
-                DELETE FROM transaction;
-                -- Create a special transaction to represent object creation.
-                INSERT INTO transaction (tid, username, description) VALUES
-                    (0, 'system', 'special transaction for object creation');
-                ALTER SEQUENCE zoid_seq START WITH 1;
-                """)
-            except:
-                conn.rollback()
-                raise
-            else:
-                conn.commit()
-        finally:
-            self.close(conn, cursor)
+        def callback(conn, cursor):
+            cursor.execute("""
+            DELETE FROM object_refs_added;
+            DELETE FROM object_ref;
+            DELETE FROM current_object;
+            DELETE FROM object_state;
+            DELETE FROM transaction;
+            -- Create a special transaction to represent object creation.
+            INSERT INTO transaction (tid, username, description) VALUES
+                (0, 'system', 'special transaction for object creation');
+            ALTER SEQUENCE zoid_seq START WITH 1;
+            """)
+        self._open_and_call(callback)
 
+    def drop_all(self):
+        """Drop all tables and sequences."""
+        def callback(conn, cursor):
+            cursor.execute("SELECT tablename FROM pg_tables")
+            existent = set([name for (name,) in cursor])
+            for tablename in ('pack_state_tid', 'pack_state',
+                    'pack_object', 'object_refs_added', 'object_ref',
+                    'current_object', 'object_state', 'transaction',
+                    'commit_lock', 'pack_lock'):
+                if tablename in existent:
+                    cursor.execute("DROP TABLE %s" % tablename)
+            cursor.execute("DROP SEQUENCE zoid_seq")
+        self._open_and_call(callback)
 
     def open(self,
             isolation=psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED):
@@ -255,12 +251,10 @@ class PostgreSQLAdapter(Adapter):
 
     def get_db_size(self):
         """Returns the approximate size of the database in bytes"""
-        conn, cursor = self.open()
-        try:
+        def callback(conn, cursor):
             cursor.execute("SELECT pg_database_size(current_database())")
             return cursor.fetchone()[0]
-        finally:
-            self.close(conn, cursor)
+        return self._open_and_call(callback)
 
     def get_current_tid(self, cursor, oid):
         """Returns the current integer tid for an object.
