@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2008 Zope Corporation and Contributors.
+# Copyright (c) 2008 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -52,7 +52,7 @@ class RelStorage(BaseStorage,
     """Storage to a relational database, based on invalidation polling"""
 
     def __init__(self, adapter, name=None, create=True,
-            read_only=False, options=None):
+            read_only=False, options=None, **kwoptions):
         if name is None:
             name = 'RelStorage on %s' % adapter.__class__.__name__
 
@@ -61,6 +61,14 @@ class RelStorage(BaseStorage,
         self._is_read_only = read_only
         if options is None:
             options = Options()
+            for key, value in kwoptions.iteritems():
+                if key in options.__dict__:
+                    setattr(options, key, value)
+                else:
+                    raise TypeError("Unknown parameter: %s" % key)
+        elif kwoptions:
+            raise TypeError("The RelStorage constructor accepts either "
+                "an options parameter or keyword arguments, not both")
         self._options = options
         self._cache_client = None
 
@@ -238,7 +246,9 @@ class RelStorage(BaseStorage,
     def _get_oid_cache_key(self, oid_int):
         """Return the cache key for finding the current tid.
 
-        This is overridden by BoundRelStorage.
+        This is overridden by BoundRelStorage.  This version always returns
+        None because a non-bound storage does not have a prev_polled_tid,
+        which is required for cache invalidation.
         """
         return None
 
@@ -261,12 +271,18 @@ class RelStorage(BaseStorage,
         else:
             msg.append("Current transaction is %d" % row[0])
 
-        rows = adapter.iter_object_history(cursor, oid_int)
         tids = []
-        for row in rows:
-            tids.append(row[0])
-            if len(tids) >= 10:
-                break
+        try:
+            rows = adapter.iter_object_history(cursor, oid_int)
+        except KeyError:
+            # The object has no history, at least from the point of view
+            # of the current database load connection.
+            pass
+        else:
+            for row in rows:
+                tids.append(row[0])
+                if len(tids) >= 10:
+                    break
         msg.append("Recent object tids: %s" % repr(tids))
         log.warning('; '.join(msg))
 
@@ -648,6 +664,15 @@ class RelStorage(BaseStorage,
         txn = self._prepared_txn
         assert txn is not None
         self._adapter.commit_phase2(self._store_cursor, txn)
+        cache = self._cache_client
+        if cache is not None:
+            if cache.incr('commit_count') is None:
+                # Use the current time as an initial commit_count value.
+                cache.add('commit_count', int(time.time()))
+                # A concurrent committer could have won the race to set the
+                # initial commit_count.  Increment commit_count so that it
+                # doesn't matter who won.
+                cache.incr('commit_count')
         self._prepared_txn = None
         self._ltid = self._tid
         self._tid = None
@@ -804,7 +829,7 @@ class RelStorage(BaseStorage,
             self._lock_release()
 
 
-    def pack(self, t, referencesf):
+    def pack(self, t, referencesf, sleep=time.sleep):
         if self._is_read_only:
             raise POSException.ReadOnlyError()
 
@@ -831,20 +856,28 @@ class RelStorage(BaseStorage,
                 # Find the latest commit before or at the pack time.
                 tid_int = adapter.choose_pack_transaction(pack_point_int)
                 if tid_int is None:
-                    # Nothing needs to be packed.
+                    log.debug("all transactions before %s have already "
+                        "been packed", time.ctime(t))
                     return
 
+                if self._options.pack_dry_run:
+                    log.info("pack: beginning dry run")
+
                 s = time.ctime(TimeStamp(p64(tid_int)).timeTime())
-                log.info("packing transactions committed %s or before", s)
+                log.info("pack: analyzing transactions committed "
+                    "%s or before", s)
 
                 # In pre_pack, the adapter fills tables with
                 # information about what to pack.  The adapter
-                # should not actually pack anything yet.
+                # must not actually pack anything yet.
                 adapter.pre_pack(tid_int, get_references, self._options)
 
-                # Now pack.
-                adapter.pack(tid_int, self._options)
-                self._after_pack()
+                if self._options.pack_dry_run:
+                    log.info("pack: dry run complete")
+                else:
+                    # Now pack.
+                    adapter.pack(tid_int, self._options, sleep=sleep)
+                    self._after_pack()
             finally:
                 adapter.release_pack_lock(lock_cursor)
         finally:
@@ -877,6 +910,10 @@ class BoundRelStorage(RelStorage):
             options=parent._options)
         # _prev_polled_tid contains the tid at the previous poll
         self._prev_polled_tid = None
+        # _commit_count contains the last polled value of the
+        # 'commit_count' cache key
+        self._commit_count = 0
+        # _poll_at is the time to poll regardless of commit_count
         self._poll_at = 0
 
     def _get_oid_cache_key(self, oid_int):
@@ -901,6 +938,30 @@ class BoundRelStorage(RelStorage):
         finally:
             self._lock_release()
 
+    def need_poll(self):
+        """Return true if polling is needed"""
+        now = time.time()
+
+        cache = self._cache_client
+        if cache is not None:
+            new_commit_count = cache.get('commit_count')
+            if new_commit_count != self._commit_count:
+                # There is new data ready to poll
+                self._commit_count = new_commit_count
+                self._poll_at = now
+                return True
+
+        if not self._load_transaction_open:
+            # Since the load connection is closed or does not have
+            # a transaction in progress, polling is required.
+            return True
+
+        if now >= self._poll_at:
+            # The poll timeout has expired
+            return True
+
+        return False
+
     def poll_invalidations(self):
         """Looks for OIDs of objects that changed since _prev_polled_tid
 
@@ -914,14 +975,10 @@ class BoundRelStorage(RelStorage):
                 return {}
 
             if self._options.poll_interval:
-                now = time.time()
-                if self._load_transaction_open and now < self._poll_at:
-                    # It's not yet time to poll again.  The previous load
-                    # transaction is still open, so it's safe to
-                    # ignore this poll.
+                if not self.need_poll():
                     return {}
-                # else poll now after resetting the timeout
-                self._poll_at = now + self._options.poll_interval
+                # reset the timeout
+                self._poll_at = time.time() + self._options.poll_interval
 
             self._restart_load()
             conn = self._load_conn
@@ -1076,10 +1133,20 @@ class Record(DataRecord):
 
 
 class Options:
-    """Options for tuning RelStorage."""
+    """Options for tuning RelStorage.
+
+    These parameters can be provided as keyword options in the RelStorage
+    constructor.  For example:
+
+        storage = RelStorage(adapter, pack_gc=True, pack_dry_run=True)
+
+    Alternatively, the RelStorage constructor accepts an options
+    parameter, which should be an Options instance.
+    """
     def __init__(self):
         self.poll_interval = 0
         self.pack_gc = True
+        self.pack_dry_run = False
         self.pack_batch_timeout = 5.0
         self.pack_duty_cycle = 0.5
         self.pack_max_delay = 20.0
